# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Wraps Designer as a gradient-free optimizer."""

from typing import Callable, List, TypeVar, Sequence

from absl import logging
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.algorithms.core import abstractions
from vizier._src.algorithms.optimizers import base

_Features = TypeVar('_Features')


class DesignerAsOptimizer(base.GradientFreeOptimizer):
  """Wraps a Designer into GradientFreeOptimizer."""

  def __init__(self,
               designer_factory: Callable[[vz.ProblemStatement],
                                          abstractions.Designer],
               *,
               batch_size: int = 100,
               num_evaluations: int = 15000):
    """Init.

    Args:
      designer_factory:
      batch_size: In each iteration, ask the designer to generate this many
        candidate trials.
      num_evaluations: Total number of trials to be evaluated on `score_fn`.
    """
    self._designer_factory = designer_factory
    self._batch_size = batch_size
    self._num_evaluations = num_evaluations

  def optimize(
      self,
      score_fn: base.BatchTrialScoreFunction,
      problem: vz.ProblemStatement,
      *,
      count: int = 1,
      budget_factor: float = 1.0,
      seed_candidates: Sequence[vz.TrialSuggestion] = tuple(),
  ) -> List[vz.Trial]:
    # Use the in-ram supporter as a pseudo-client for running a study in RAM.
    study = pythia.InRamPolicySupporter(problem)

    # Save the designer for debugging purposes only.
    self._designer = self._designer_factory(problem)
    num_iterations = max(
        int(self._num_evaluations * budget_factor) // self._batch_size, 1)
    logging.info(
        'Optimizing the acquisition for %s iterations of %s trials each',
        num_iterations, self._batch_size)

    for _ in range(num_iterations):
      trials = study.AddSuggestions(self._designer.suggest(self._batch_size))
      if not trials:
        break
      scores = score_fn(trials)
      # Check that scores are (N, 1) arrays as in BatchTrialScoreFunction.
      for k, v in scores.items():
        if v.shape != (len(trials), 1):
          raise ValueError(
              f'Incorrect shape {v.shape} in scores {scores[k]}\n'
              f'Expected shape is {(len(trials), 1)}'
          )
      for i, trial in enumerate(trials):
        # TODO: Decide what to do with NaNs scores.
        trial.complete(
            vz.Measurement({k: v[i].item() for k, v in scores.items()}))
      self._designer.update(
          abstractions.CompletedTrials(trials), abstractions.ActiveTrials()
      )
    logging.info(
        'Finished running the optimization study. Extracting the best trials...'
    )
    return study.GetBestTrials(count=count)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for scheduled designers."""

from typing import Sequence

import attrs
import jax
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.algorithms.designers import gp_bandit
from vizier._src.algorithms.designers import gp_ucb_pe
from vizier._src.algorithms.designers import random
from vizier._src.algorithms.designers import scheduled_designer
from vizier._src.algorithms.designers import scheduled_gp_bandit
from vizier._src.algorithms.designers import scheduled_gp_ucb_pe
from vizier._src.algorithms.testing import test_runners
from vizier.testing import test_studies

from absl.testing import absltest


@attrs.define
class MockParameterizedDesigner(vza.Designer):
  """Mock parameterized designer."""

  problem: vz.ProblemStatement
  parameter1: float = 0.0
  parameter2: float = 0.0

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    """Update the designer based on completed and pending trials."""
    del completed
    del all_active

  def suggest(self, count: int = 1) -> Sequence[vz.TrialSuggestion]:
    """Suggest trials."""
    return random.RandomDesigner(self.problem.search_space).suggest(count)


class DirectDesignerStateUpdater(scheduled_designer.DesignerStateUpdater):
  """Direct designer state updater."""

  def __call__(
      self, designer: vza.Designer, params: scheduled_designer.ParamsValues
  ) -> None:
    """Update the designer state assuming the params are class attributes."""
    for param_name, param_value in params.items():
      setattr(designer, param_name, param_value)


class ScheduledDesignerTest(absltest.TestCase):

  def test_schedule_designer(self):
    expected_total_num_trials = 10
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name="metric", goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    param1 = scheduled_designer.LinearScheduledParam(
        init_value=10.5, final_value=2.1
    )
    param2 = scheduled_designer.ExponentialScheduledParam(
        init_value=1.5, final_value=20.1, rate=1.7
    )
    mock_scheduled_designer = scheduled_designer.ScheduledDesigner(
        problem,
        designer_factory=MockParameterizedDesigner,
        designer_state_updater=DirectDesignerStateUpdater(),
        scheduled_params={"parameter1": param1, "parameter2": param2},
        expected_total_num_trials=expected_total_num_trials,
    )
    # Check initial values.
    self.assertEqual(mock_scheduled_designer.designer.parameter1, 10.5)  # pytype: disable=attribute-error
    self.assertEqual(mock_scheduled_designer.designer.parameter2, 1.5)  # pytype: disable=attribute-error
    # Check suggestions.
    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=expected_total_num_trials,
            batch_size=1,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(mock_scheduled_designer),
        expected_total_num_trials,
    )
    # Check final values.
    self.assertAlmostEqual(mock_scheduled_designer.designer.parameter1, 2.1)  # pytype: disable=attribute-error
    self.assertAlmostEqual(mock_scheduled_designer.designer.parameter2, 20.1)  # pytype: disable=attribute-error

  def test_validate_suggested_num_trials(self):
    # Test that updating the designer with trials changes the state.
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name="metric", goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    param1 = scheduled_designer.LinearScheduledParam(
        init_value=10.5, final_value=2.1
    )
    param2 = scheduled_designer.ExponentialScheduledParam(
        init_value=1.5, final_value=20.1, rate=1.7
    )
    mock_scheduled_designer = scheduled_designer.ScheduledDesigner(
        problem,
        designer_factory=MockParameterizedDesigner,
        designer_state_updater=DirectDesignerStateUpdater(),
        scheduled_params={"parameter1": param1, "parameter2": param2},
        expected_total_num_trials=10,
    )
    # Generate active and completed trials.
    active_trials = test_studies.flat_continuous_space_with_scaling_trials(2)
    completed_trials = [
        s.to_trial()
        for s in test_studies.flat_continuous_space_with_scaling_trials(4)
    ]
    for trial in completed_trials:
      trial.complete(vz.Measurement({"metric": 1.2}), inplace=True)
    # Update the scheduled designer.
    mock_scheduled_designer.update(
        vza.CompletedTrials(completed_trials), vza.ActiveTrials(active_trials)
    )
    # Validate that the state was updated.
    self.assertEqual(
        mock_scheduled_designer.num_incorporated_suggested_trials, 4 + 2
    )

  def test_scheduled_designer_serialization(self):
    expected_total_num_trials = 10
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name="metric", goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )
    param1 = scheduled_designer.LinearScheduledParam(
        init_value=10.5, final_value=2.1
    )
    param2 = scheduled_designer.ExponentialScheduledParam(
        init_value=1.5, final_value=20.1, rate=1.7
    )
    mock_scheduled_designer = scheduled_designer.ScheduledDesigner(
        problem,
        designer_factory=MockParameterizedDesigner,
        designer_state_updater=DirectDesignerStateUpdater(),
        scheduled_params={"parameter1": param1, "parameter2": param2},
        expected_total_num_trials=expected_total_num_trials,
    )
    # Making several suggestions so the state would change.
    mock_scheduled_designer.suggest(count=1)
    mock_scheduled_designer.suggest(count=3)
    # Store the state in metadata.
    state = mock_scheduled_designer.dump()
    # Create a new designer and load state.
    new_mock_scheduled_designer = scheduled_designer.ScheduledDesigner(
        problem,
        designer_factory=MockParameterizedDesigner,
        designer_state_updater=DirectDesignerStateUpdater(),
        scheduled_params={"parameter1": param1, "parameter2": param2},
        expected_total_num_trials=expected_total_num_trials,
    )
    new_mock_scheduled_designer.load(state)
    self.assertEqual(
        new_mock_scheduled_designer._num_incorporated_suggested_trials, 4
    )


class ScheduledGpBanditTest(absltest.TestCase):

  def test_scheduled_gp_bandit(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name="metric", goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    def _gp_bandit_factory(problem):
      return gp_bandit.VizierGPBandit(problem)

    scheduled_desinger = scheduled_gp_bandit.ScheduledGPBanditFactory(
        gp_bandit_factory=_gp_bandit_factory,
        expected_total_num_trials=2,
        init_ucb_coefficient=4.0,
        final_ucb_coefficient=1.0,
        decay_ucb_coefficient=1.2,
    )(problem)

    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=2,
            batch_size=1,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(scheduled_desinger),
        2,
    )


class ScheduledGpUcbPeTest(absltest.TestCase):

  def test_scheduled_gp_ucb_pe(self):
    problem = vz.ProblemStatement(
        test_studies.flat_continuous_space_with_scaling()
    )
    problem.metric_information.append(
        vz.MetricInformation(
            name="metric", goal=vz.ObjectiveMetricGoal.MAXIMIZE
        )
    )

    def _gp_ucb_pe_factory(
        problem: vz.ProblemStatement,
    ) -> gp_ucb_pe.VizierGPUCBPEBandit:
      return gp_ucb_pe.VizierGPUCBPEBandit(problem)

    scheduled_desinger = scheduled_gp_ucb_pe.ScheduledGPUCBPEFactory(
        gp_ucb_pe_factory=_gp_ucb_pe_factory,
        expected_total_num_trials=10,
        init_ucb_coefficient=4.0,
        final_ucb_coefficient=1.0,
        decay_ucb_coefficient=1.2,
        init_explore_region_ucb_coefficient=1.0,
        final_explore_region_ucb_coefficient=0.5,
        decay_explore_region_ucb_coefficient=1.2,
        init_ucb_overwrite_probability=0.25,
        final_ucb_overwrite_probability=0.0,
        decay_ucb_overwrite_probability=1.0,
    )(problem)

    self.assertLen(
        test_runners.RandomMetricsRunner(
            problem,
            iters=10,
            batch_size=1,
            verbose=1,
            validate_parameters=True,
            seed=1,
        ).run_designer(scheduled_desinger),
        10,
    )


if __name__ == "__main__":
  jax.config.update("jax_enable_x64", True)
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Sinusodial function for testing."""

from typing import Optional
import unittest

from absl import logging
import chex
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
import tree
from vizier._src.jax import stochastic_process_model as sp
from vizier._src.jax import types
from vizier._src.jax.optimizers import core
from vizier._src.jax.types import Bounds


tfb = tfp.bijectors


@chex.assert_max_traces(1)
def loss_fn(inputs):
  # Sinusoidal function minimized at all inputs being zero. (min value=-2)
  # There are lots of local optima between [-3, 3]
  xs = jnp.concatenate(jax.tree_util.tree_flatten(inputs)[0], axis=-1)
  xs = jax.nn.sigmoid(xs)
  return (
      jnp.mean(jnp.cos(xs * 5 * 2 * jnp.pi) * (2 - 5 * (xs - 0.5) ** 2)) + 2,
      dict(),
  )


def setup(rng):
  rng1, rng2 = jax.random.split(rng, 2)
  return {
      'x1': jax.random.uniform(rng1, shape=(1,), minval=-3, maxval=3),
      'x2': jax.random.uniform(rng2, shape=(2,), minval=-3.0, maxval=3.0),
  }


def _make_constraints(b: Optional[types.Bounds]) -> Optional[chex.ArrayTree]:
  if b is None:
    return None
  return {'x1': jnp.array([b]), 'x2': jnp.array([b, b])}


def bounds_to_constraints(
    bounds: Optional[Bounds],
    nest_constraint: bool = True,
) -> Optional[sp.Constraint]:
  if bounds is None:
    return None
  if nest_constraint:
    bounds = tree.map_structure(_make_constraints, bounds)
  return sp.Constraint.create(bounds, tfb.SoftClip)


class ConvergenceTestMixin(unittest.TestCase):
  """Mixin for subclasses."""

  def assert_converges(
      self,
      optimize: core.Optimizer,
      *,
      constraints: sp.Constraint,
      threshold: float = 5e-3,
      random_restarts: int,
  ) -> None:
    """Assert the optimizer converges."""
    rngs = jax.random.split(jax.random.PRNGKey(1), random_restarts + 1)
    optimal_params, metrics = optimize(
        init_params=jax.vmap(setup)(rngs[1:]),
        loss_fn=jax.jit(loss_fn),
        rng=rngs[0],
        constraints=constraints,
    )
    logging.info('Optimal: %s', optimal_params)

    self.assertLessEqual(loss_fn(optimal_params)[0], threshold)
    if metrics['loss'].shape[0] > 1:
      np.testing.assert_array_less(
          metrics['loss'][-1, :], metrics['loss'][0, :]
      )

    if (constraints is not None) and (constraints.bounds is not None):
      bounds = constraints.bounds
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(bounds[0])):
        if b_ is not None:
          np.testing.assert_array_less(b_, y_)
      for y_, b_ in zip(tree.flatten(optimal_params), tree.flatten(bounds[1])):
        if b_ is not None:
          np.testing.assert_array_less(y_, b_)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Categorical benchmarks from https://github.com/QUVA-Lab/COMBO.

Lines with 'ATTENTION' mean that code was modified from original Github, in
order to fix bugs with the original code.
"""

# pyformat: disable
from typing import List, Optional, Sequence, Tuple
import numpy as np

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier._src.benchmarks.experimenters.combo import common

fileOpen = open


class IsingExperimenter(experimenter.Experimenter):
  """Ising Sparisification Problem."""

  def __init__(self,
               lamda: float = 1e-2,
               ising_grid_h: int = 4,
               ising_grid_w: int = 4,
               ising_n_edges: int = 24,
               random_seed: Optional[int] = None):
    self._lamda = lamda
    self._ising_grid_h = ising_grid_h
    self._ising_grid_w = ising_grid_w
    self._ising_n_edges = ising_n_edges
    self._interaction = common.generate_ising_interaction(
        self._ising_grid_h, self._ising_grid_w, random_seed)
    self._covariance, self._partition_original = common.spin_covariance(
        self._interaction, (self._ising_grid_h, self._ising_grid_w))

    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value == 'True')
          for i in range(self._ising_n_edges)
      ])
      x_h, x_v = self._bocs_consistency_mapping(x)
      interaction_sparsified = x_h * self._interaction[
          0], x_v * self._interaction[1]
      log_partition_sparsified = common.log_partition(
          interaction_sparsified, (self._ising_grid_h, self._ising_grid_w))
      evaluation = common.ising_dense(
          ising_grid_h=self._ising_grid_h,
          interaction_original=self._interaction,
          interaction_sparsified=interaction_sparsified,
          covariance=self._covariance,
          log_partition_original=np.log(self._partition_original),
          log_partition_new=log_partition_sparsified)
      evaluation += self._lamda * float(np.sum(x))

      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._ising_n_edges):
      root.add_bool_param(name=f'x_{i}')
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _bocs_consistency_mapping(self,
                                x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    horizontal_ind = [0, 2, 4, 7, 9, 11, 14, 16, 18, 21, 22, 23]
    vertical_ind = sorted(
        [elm for elm in range(24) if elm not in horizontal_ind])
    return x[horizontal_ind].reshape(
        (self._ising_grid_h, self._ising_grid_w - 1)), x[vertical_ind].reshape(
            (self._ising_grid_h - 1, self._ising_grid_w))


class ContaminationExperimenter(experimenter.Experimenter):
  """Contamination Control Problem."""

  def __init__(self,
               lamda: float = 1e-2,
               contamination_n_stages: int = 25,
               random_seed: Optional[int] = None):
    self._lamda = lamda
    self._contamination_n_stages = contamination_n_stages
    self._init_z, self._lambdas, self._gammas = self._generate_contamination_dynamics(
        random_seed)
    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value == 'True')
          for i in range(self._contamination_n_stages)
      ])
      evaluation = self._contamination(
          x=x,
          cost=np.ones(x.size),
          init_z=self._init_z,
          lambdas=self._lambdas,
          gammas=self._gammas,
          u=0.1,
          epsilon=0.05)
      evaluation += self._lamda * float(np.sum(x))
      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._contamination_n_stages):
      root.add_bool_param(name=f'x_{i}')
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _contamination(self, x: np.ndarray, cost: np.ndarray, init_z: np.ndarray,
                     lambdas: np.ndarray, gammas: np.ndarray, u: float,
                     epsilon: float) -> float:
    assert x.size == self._contamination_n_stages

    rho = 1.0
    n_simulations = 100

    z = np.zeros((x.size, n_simulations))
    z[0] = lambdas[0] * (1.0 - x[0]) * (1.0 - init_z) + (
        1.0 - gammas[0] * x[0]) * init_z
    for i in range(1, self._contamination_n_stages):
      z[i] = lambdas[i] * (1.0 - x[i]) * (1.0 - z[i - 1]) + (
          1.0 - gammas[i] * x[i]) * z[i - 1]

    below_threshold = z < u
    constraints = np.mean(below_threshold, axis=1) - (1.0 - epsilon)

    return np.sum(x * cost - rho * constraints)

  def _generate_contamination_dynamics(
      self, random_seed=None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    n_stages = self._contamination_n_stages
    n_simulations = 100

    init_alpha = 1.0
    init_beta = 30.0
    contam_alpha = 1.0
    contam_beta = 17.0 / 3.0
    restore_alpha = 1.0
    restore_beta = 3.0 / 7.0
    init_z = np.random.RandomState(random_seed).beta(
        init_alpha, init_beta, size=(n_simulations,))
    lambdas = np.random.RandomState(random_seed).beta(
        contam_alpha, contam_beta, size=(n_stages, n_simulations))
    gammas = np.random.RandomState(random_seed).beta(
        restore_alpha, restore_beta, size=(n_stages, n_simulations))

    return init_z, lambdas, gammas


class CentroidExperimenter(experimenter.Experimenter):
  """General parameter variant of the Ising Sparisification Problem."""

  def __init__(self,
               centroid_n_choice=3,
               centroid_grid=(4, 4),
               ising_grid_h: int = 4,
               random_seed: Optional[int] = None):
    self._centroid_n_choice = centroid_n_choice
    self._centroid_grid = centroid_grid
    self._centroid_n_edges = centroid_grid[0] * (centroid_grid[1] - 1) + (
        centroid_grid[0] - 1) * centroid_grid[1]
    self._ising_grid_h = ising_grid_h

    self._interaction_list = []
    self._covariance_list = []
    self._partition_original_list = []
    self._n_ising_models = 3
    ising_seeds = np.random.RandomState(random_seed).randint(
        0, 10000, (self._n_ising_models,))
    for i in range(self._n_ising_models):
      interaction = common.generate_ising_interaction(self._centroid_grid[0],
                                                      self._centroid_grid[1],
                                                      ising_seeds[i])
      covariance, partition_original = common.spin_covariance(
          interaction, self._centroid_grid)
      self._interaction_list.append(interaction)
      self._covariance_list.append(covariance)
      self._partition_original_list.append(partition_original)

    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value)
          for i in range(self._centroid_n_edges)
      ])
      interaction_mixed = self._edge_choice(x, self._interaction_list)
      log_partition_mixed = common.log_partition(
          interaction_mixed, self._centroid_grid)  # ATTENTION
      kld_sum = 0

      for i in range(self._n_ising_models):
        kld = common.ising_dense(
            ising_grid_h=self._centroid_grid[0],  # ATTENTION
            interaction_original=self._interaction_list[i],
            interaction_sparsified=interaction_mixed,
            covariance=self._covariance_list[i],
            log_partition_original=np.log(
                self._partition_original_list[i]),  # ATTENTION
            log_partition_new=log_partition_mixed)  # ATTENTION
        kld_sum += kld
      evaluation = float(kld_sum / float(self._n_ising_models))

      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._centroid_n_edges):
      root.add_categorical_param(
          name='x_{}'.format(i),
          feasible_values=[str(j) for j in range(self._centroid_n_choice)])
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _edge_choice(
      self, x: np.ndarray, interaction_list: List[Tuple[np.ndarray, np.ndarray]]
  ) -> Tuple[np.ndarray, np.ndarray]:
    edge_weight = np.zeros(x.shape)
    for i in range(len(interaction_list)):
      edge_weight[x == i] = np.hstack([
          interaction_list[i][0].reshape(-1), interaction_list[i][1].reshape(-1)
      ])[x == i]
    grid_h, grid_w = self._centroid_grid
    split_ind = grid_h * (grid_w - 1)
    return edge_weight[:split_ind].reshape(
        (grid_h, grid_w - 1)), edge_weight[split_ind:].reshape(
            (grid_h - 1, grid_w))


class PestControlExperimenter(experimenter.Experimenter):
  """Pest Control Problem."""

  def __init__(self,
               pest_control_n_choice: int = 5,
               pest_control_n_stages: int = 25,
               random_seed: Optional[int] = None):
    self._pest_control_n_choice = pest_control_n_choice
    self._pest_control_n_stages = pest_control_n_stages
    self._random_seed = random_seed
    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      x = np.array([
          int(suggestion.parameters[f'x_{i}'].value)
          for i in range(self._pest_control_n_stages)
      ])
      evaluation = self._pest_control_score(x)
      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._pest_control_n_stages):
      root.add_categorical_param(
          name='x_{}'.format(i),
          feasible_values=[str(j) for j in range(self._pest_control_n_choice)])
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement

  def _pest_spread(self, curr_pest_frac: float, spread_rate: float,
                   control_rate: float, apply_control: bool):
    if apply_control:
      next_pest_frac = (1.0 - control_rate) * curr_pest_frac
    else:
      next_pest_frac = spread_rate * (1 - curr_pest_frac) + curr_pest_frac
    return next_pest_frac

  def _pest_control_score(self, x: np.ndarray) -> float:
    u = 0.1
    n_stages = x.size
    n_simulations = 100

    init_pest_frac_alpha = 1.0
    init_pest_frac_beta = 30.0
    spread_alpha = 1.0
    spread_beta = 17.0 / 3.0

    control_alpha = 1.0
    control_price_max_discount = {1: 0.2, 2: 0.3, 3: 0.3, 4: 0.0}
    tolerance_develop_rate = {
        1: 1.0 / 7.0,
        2: 2.5 / 7.0,
        3: 2.0 / 7.0,
        4: 0.5 / 7.0
    }
    control_price = {1: 1.0, 2: 0.8, 3: 0.7, 4: 0.5}
    # below two changes over stages according to x
    control_beta = {1: 2.0 / 7.0, 2: 3.0 / 7.0, 3: 3.0 / 7.0, 4: 5.0 / 7.0}

    payed_price_sum = 0
    above_threshold = 0

    init_pest_frac = np.random.RandomState(self._random_seed).beta(
        init_pest_frac_alpha, init_pest_frac_beta, size=(n_simulations,))
    curr_pest_frac = init_pest_frac
    for i in range(n_stages):
      spread_rate = np.random.RandomState(self._random_seed).beta(
          spread_alpha, spread_beta, size=(n_simulations,))
      do_control = x[i] > 0
      if do_control:
        control_rate = np.random.RandomState(self._random_seed).beta(
            control_alpha, control_beta[x[i]], size=(n_simulations,))
        next_pest_frac = self._pest_spread(curr_pest_frac, spread_rate,
                                           control_rate, True)
        # Tolerance has been developed for pesticide type 1.
        control_beta[x[i]] += tolerance_develop_rate[x[i]] / float(n_stages)
        # You will get a discount.
        payed_price = control_price[x[i]] * (
            1.0 - control_price_max_discount[x[i]] / float(n_stages) *
            float(np.sum(x == x[i])))
      else:
        next_pest_frac = self._pest_spread(curr_pest_frac, spread_rate, 0,
                                           False)
        payed_price = 0
      payed_price_sum += payed_price
      above_threshold += np.mean(curr_pest_frac > u)
      curr_pest_frac = next_pest_frac

    return payed_price_sum + above_threshold


# MAXSAT Files can be found in
# https://github.com/QUVA-Lab/COMBO/tree/master/COMBO/experiments/MaxSAT/maxsat2018_data.

MAXSAT28_FILE = 'maxsat2018_data/maxcut-johnson8-2-4.clq.wcnf'
MAXSAT43_FILE = 'maxsat2018_data/maxcut-hamming8-2.clq.wcnf'
MAXSAT60_FILE = 'maxsat2018_data/frb-frb10-6-4.wcnf'


class MAXSATExperimenter(experimenter.Experimenter):
  """MAXSAT Problem."""

  def __init__(self, data_filename: str):
    self._data_filename = data_filename
    f = fileOpen(self._data_filename, 'rt')
    line_str = f.readline()
    while line_str[:2] != 'p ':
      line_str = f.readline()
    self._n_variables = int(line_str.split(' ')[2])
    self._n_clauses = int(line_str.split(' ')[3])
    self._n_vertices = np.array([2] * self._n_variables)
    raw_clauses = [(float(clause_str.split(' ')[0]),
                    clause_str.split(' ')[1:-1])
                   for clause_str in f.readlines()]
    f.close()
    weights = np.array([elm[0] for elm in raw_clauses]).astype(np.float32)
    weight_mean = np.mean(weights)
    weight_std = np.std(weights)
    self._weights = (weights - weight_mean) / weight_std
    self._clauses = []
    for _, clause in raw_clauses:
      pair = ([abs(int(elm)) - 1 for elm in clause],
              [int(elm) > 0 for elm in clause])
      self._clauses.append(pair)

    self._problem_statement = self.problem_statement()

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]):
    for suggestion in suggestions:
      # TODO: Switch to using StudyConfig.
      bools = [
          int(suggestion.parameters[f'x_{i}'].value == 'True')
          for i in range(self._n_variables)
      ]
      x = np.array(bools, dtype=bool)
      satisfied = np.array([
          (x[clause[0]] == clause[1]).any() for clause in self._clauses
      ])
      evaluation = -np.sum(self._weights * satisfied)
      suggestion.complete(
          pyvizier.Measurement(metrics={
              self._problem_statement.single_objective_metric_name: evaluation
          }))

  def problem_statement(self) -> pyvizier.ProblemStatement:
    problem_statement = pyvizier.ProblemStatement()
    root = problem_statement.search_space.root
    for i in range(self._n_variables):
      root.add_bool_param(name=f'x_{i}')
    problem_statement.metric_information.append(
        pyvizier.MetricInformation(
            name='main_objective', goal=pyvizier.ObjectiveMetricGoal.MINIMIZE))
    return problem_statement



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for failing."""

from vizier import pyvizier as vz
from vizier._src.algorithms.testing import failing
from absl.testing import absltest


class FailingTest(absltest.TestCase):

  def test_failing_designer(self):
    failing_designer = failing.FailingDesigner()
    with self.assertRaises(failing.FailedSuggestError):
      failing_designer.suggest(1)

  def test_alternate_failing_designer(self):
    search_space = vz.SearchSpace()
    search_space.root.add_float_param("x", 0.0, 1.0)
    alt_failing_designer = failing.AlternateFailingDesigner(search_space)
    for _ in range(5):
      alt_failing_designer.suggest(1)
      with self.assertRaises(failing.FailedSuggestError):
        alt_failing_designer.suggest(1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""RPC functions implemented from vizier_service.proto."""

import collections
import datetime
import threading
from typing import Optional

from absl import logging
import grpc
import numpy as np
import sqlalchemy as sqla
from vizier import pythia
from vizier import pyvizier as vz
from vizier._src.service import constants
from vizier._src.service import custom_errors
from vizier._src.service import grpc_util
from vizier._src.service import pythia_service
from vizier._src.service import ram_datastore
from vizier._src.service import resources
from vizier._src.service import sql_datastore
from vizier._src.service import stubs_util
from vizier._src.service import study_pb2
from vizier._src.service import types
from vizier._src.service import vizier_oss_pb2
from vizier._src.service import vizier_service_pb2
from vizier._src.service import vizier_service_pb2_grpc
from vizier.service import pyvizier as svz

from google.longrunning import operations_pb2
from google.protobuf import empty_pb2
from google.protobuf import timestamp_pb2
from google.rpc import code_pb2
from google.rpc import status_pb2


def _get_current_time() -> timestamp_pb2.Timestamp:
  now = timestamp_pb2.Timestamp()
  now.GetCurrentTime()
  return now


StudyResource = resources.StudyResource
TrialResource = resources.TrialResource


# TODO: remove context = None
# TODO: remove context = None
class VizierServicer(vizier_service_pb2_grpc.VizierServiceServicer):
  """Implements the GRPC functions outlined in vizier_service.proto."""

  # Trial states in which a trial can be modified.
  _TRIAL_MUTABLE_STATES = (
      study_pb2.Trial.State.ACTIVE,
      study_pb2.Trial.State.STOPPING,
  )

  def __init__(
      self,
      database_url: Optional[str] = constants.SQL_LOCAL_URL,
      early_stop_recycle_period: datetime.timedelta = datetime.timedelta(
          seconds=60
      ),
      default_pythia_service: Optional[types.PythiaService] = None,
  ):
    """Initializes the service.

    Creates the datastore and relevant locks for multhreading. Note that the
    datastore input/output is assumed to always be pass-by-value.

    Args:
      database_url: URL to the SQL database. If None, it connects to our custom
        RAM Datastore.
      early_stop_recycle_period: Amount of time needed to pass before recycling
        an early stopping operation. See `CheckEarlyStoppingState` for more
        details.
      default_pythia_service: Default PythiaService to use when
        StudyConfig.pythia_endpoint is unset. If None, creates a local
        PythiaServicer.
    """
    # By default, uses a local PythiaServicer instance.
    self.default_pythia_service: types.PythiaService = (
        default_pythia_service
        or pythia_service.PythiaServicer(vizier_service=self)
    )

    if database_url is None:
      self.datastore = ram_datastore.NestedDictRAMDataStore()
    else:
      engine = sqla.create_engine(
          database_url,
          connect_args={'check_same_thread': False},
          echo=False,  # Set True to log transactions for debugging.
          future=True,  # Backward compatibility with sqlalchemy 1.4.
          poolclass=sqla.pool.StaticPool,
      )
      self.datastore = sql_datastore.SQLDataStore(engine)

    # For database edits using owner names.
    self._owner_name_to_lock = collections.defaultdict(threading.Lock)
    # For database edits using study names.
    self._study_name_to_lock = collections.defaultdict(threading.Lock)
    # For calls to Pythia (SuggestTrials and CheckTrialEarlyStoppingState).
    self._operation_lock = collections.defaultdict(threading.Lock)

    self._early_stop_recycle_period = early_stop_recycle_period

  def _select_pythia_service(
      self, endpoint: Optional[str] = None
  ) -> types.PythiaService:
    """Selects PythiaService to use."""
    # TODO: Add test for StudyConfig endpointing.
    if endpoint is None:
      logging.info('Using default PythiaServicer.')
      return self.default_pythia_service

    logging.info('Connecting to Pythia endpoint: %s', endpoint)
    pythia_stub = stubs_util.create_pythia_server_stub(endpoint)
    logging.info('Created Pythia Server stub: %s', pythia_stub)
    return pythia_stub

  def _study_is_immutable(self, study_name: str) -> bool:
    """Checks if study is immutable to block study-related mutations."""
    study = self.datastore.load_study(study_name)
    return study.state not in (
        study_pb2.Study.State.ACTIVE,
        study_pb2.Study.State.STATE_UNSPECIFIED,
    )

  def CreateStudy(
      self,
      request: vizier_service_pb2.CreateStudyRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Study:
    """Creates a study or loads an existing one.

    The initial request contains the study without the study name, only
    a user/client specified display_name for locating a potential pre-existing
    study. The returned study will then have the service-provided resource name.

    Args:
      request: A CreateStudyRequest.
      context: Optional GRPC ServicerContext.

    Returns:
      study: The Study created with generated service resource name.
    """
    study = request.study
    owner_id = resources.OwnerResource.from_name(request.parent).owner_id
    if request.study.name:
      e = ValueError(
          'Maximum number of studies reached for owner {}.'.format(owner_id)
      )
      grpc_util.handle_exception(e, context)
    if not request.study.display_name:
      grpc_util.handle_exception(
          ValueError('Study display_name must be specified.'), context
      )

    with self._owner_name_to_lock[request.parent]:
      # Database creates a new active study or loads existing study using the
      # display name.
      try:
        possible_candidate_studies = self.datastore.list_studies(request.parent)
      except custom_errors.NotFoundError:
        possible_candidate_studies = []

      # Check if all possible study_id's have been taken.
      if len(possible_candidate_studies) >= constants.MAX_STUDY_ID:
        e = ValueError(
            'Maximum number of studies reached for owner {}.'.format(owner_id)
        )
        grpc_util.handle_exception(e, context)

      for candidate_study in possible_candidate_studies:
        if candidate_study.display_name == request.study.display_name:
          logging.info(
              'Found existing study of owner=%s display_name=%s!',
              owner_id,
              candidate_study.display_name,
          )
          return candidate_study

      # No study in the database matches the resource name. Making a new one.
      # study_id must be unique among existing studies.
      study_id = study.display_name

      # Finally create study in database and return it.
      study.name = StudyResource(owner_id, study_id).name
      self.datastore.create_study(study)
    return study

  def GetStudy(
      self,
      request: vizier_service_pb2.GetStudyRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Study:
    """Gets a Study by name. If the study does not exist, return error."""
    return self.datastore.load_study(request.name)

  def ListStudies(
      self,
      request: vizier_service_pb2.ListStudiesRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> vizier_service_pb2.ListStudiesResponse:
    """Lists all the studies in a region for an associated project."""
    studies = self.datastore.list_studies(request.parent)
    return vizier_service_pb2.ListStudiesResponse(studies=studies)

  def DeleteStudy(
      self,
      request: vizier_service_pb2.DeleteStudyRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> empty_pb2.Empty:
    """Deletes a Study."""
    self.datastore.delete_study(request.name)
    return empty_pb2.Empty()

  def SetStudyState(
      self,
      request: vizier_service_pb2.SetStudyStateRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Study:
    with self._study_name_to_lock[request.parent]:
      study = self.datastore.load_study(request.parent)
      study.state = request.state
      self.datastore.update_study(study)
    return study

  def SuggestTrials(
      self,
      request: vizier_service_pb2.SuggestTrialsRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> operations_pb2.Operation:
    """Adds one or more Trials to a Study, with parameter values suggested by a Pythia policy.

    The logic is as follows:
    1. If there is already an active (not done) operation, simply return that.
    2. Else, create a new active operation.
    3. We need the requested number of ACTIVE trials to return. These will come
    from 3 sources:
      A. ACTIVE trials already assigned to the client.
      B. REQUESTED trials which not been assigned to any client.
      C. Pythia-computed suggestions.

    We first look from source A, then source B, then source C. If we ever reach
    source C and Pythia over-delivers too many suggestions (e.g. evolutionary
    algorithms sometimes do so because of batched behaviors), we put the extra
    suggestions into the REQUESTED pool (i.e. source B) for future use.

    The returned operation can either be:
    1. Done (Successfully contains the requested number of suggestions)
    2. Error-ed, which will happen if Pythia had an issue producing the right
    number of trials (via numerical crash or under-delivering).

    Args:
      request:
      context:

    Returns:
      A long-running operation associated with the generation of Trial
      suggestions. When this long-running operation succeeds, it will contain a
      [SuggestTrialsResponse].

    Raises:
      ImmutableStudyError: If study was already immutable.
    """
    # Convenient names and id's to be used below.
    study_name = request.parent
    if self._study_is_immutable(study_name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot suggest trial.'.format(study_name)
      )
      grpc_util.handle_exception(e, context)

    study_resource = StudyResource.from_name(study_name)
    study_id = study_resource.study_id
    owner_id = study_resource.owner_id

    # Don't allow simultaneous SuggestTrial or EarlyStopping calls to be
    # processed.
    with self._operation_lock[request.parent]:
      study = self.datastore.load_study(request.parent)

      # Checks for a non-done operation in the database with this name.
      active_op_filter_fn = lambda op: not op.done
      try:
        active_op_list = self.datastore.list_suggestion_operations(
            study_name, request.client_id, active_op_filter_fn
        )
      except custom_errors.NotFoundError:
        active_op_list = []
      if active_op_list:
        return active_op_list[0]  # We've found the active one!

      start_time = _get_current_time()
      # Create a new Op if there aren't any active (not done) ops.
      try:
        old_op_number = self.datastore.max_suggestion_operation_number(
            study_name, request.client_id
        )
      except custom_errors.NotFoundError:
        old_op_number = 0
      new_op_number = old_op_number + 1
      new_op_name = resources.SuggestionOperationResource(
          owner_id, study_id, request.client_id, new_op_number
      ).name
      output_op = operations_pb2.Operation(name=new_op_name, done=False)
      self.datastore.create_suggestion_operation(output_op)

      # Check how many ACTIVE trials already exist for this client only.
      all_trials = self.datastore.list_trials(study_name)
      active_trials = [
          t
          for t in all_trials
          if t.state == study_pb2.Trial.State.ACTIVE
          and t.client_id == request.client_id
      ]
      if len(active_trials) >= request.suggestion_count:
        output_op.response.value = vizier_service_pb2.SuggestTrialsResponse(
            trials=active_trials[: request.suggestion_count],
            start_time=start_time,
        ).SerializeToString()
        output_op.done = True
        self.datastore.update_suggestion_operation(output_op)
        return output_op

      # Get suggestions from the pool of requested trials.
      output_trials = active_trials
      requested_trials = [
          t for t in all_trials if t.state == study_pb2.Trial.State.REQUESTED
      ]
      while requested_trials and request.suggestion_count > len(output_trials):
        assigned_trial = requested_trials.pop()
        assigned_trial.state = study_pb2.Trial.State.ACTIVE
        assigned_trial.client_id = request.client_id
        assigned_trial.start_time.CopyFrom(start_time)
        self.datastore.update_trial(assigned_trial)
        output_trials.append(assigned_trial)

      if len(output_trials) == request.suggestion_count:
        # We've finished collecting enough trials from the REQUESTED pool.
        output_op.response.value = vizier_service_pb2.SuggestTrialsResponse(
            trials=output_trials, start_time=start_time
        ).SerializeToString()
        output_op.done = True
        self.datastore.update_suggestion_operation(output_op)
        return output_op

      # Still need more suggestions. Pythia begins computing missing amount.
      study_config = svz.StudyConfig.from_proto(study.study_spec)
      study_descriptor = vz.StudyDescriptor(
          config=study_config,
          guid=study_name,
          max_trial_id=self.datastore.max_trial_id(study_name),
      )
      suggest_request = pythia.SuggestRequest(
          study_descriptor=study_descriptor,
          count=request.suggestion_count - len(output_trials),
      )

      # Convert request, send to Pythia, and obtain suggestions.
      suggest_request_proto = svz.SuggestConverter.to_request_proto(
          suggest_request
      )
      suggest_request_proto.algorithm = study.study_spec.algorithm

      try:
        temp_pythia_service = self._select_pythia_service(
            study_config.pythia_endpoint
        )
        suggest_decision_proto = temp_pythia_service.Suggest(
            suggest_request_proto
        )
      # Pythia can raise any exception, captured inside grpc.RpcError.
      except grpc.RpcError as e:
        output_op.error.CopyFrom(
            status_pb2.Status(code=code_pb2.Code.INTERNAL, message=str(e))
        )
        logging.exception(
            'Failed to request trials from Pythia for request: %s', request
        )
        output_op.done = True
        self.datastore.update_suggestion_operation(output_op)
        return output_op

      # Check if we received enough suggestions.
      if len(
          suggest_decision_proto.suggestions
      ) < request.suggestion_count - len(output_trials):
        logging.warning(
            'Requested at least %d suggestions but Pythia only produced %d.',
            request.suggestion_count - len(output_trials),
            len(suggest_decision_proto.suggestions),
        )
      suggest_decision = svz.SuggestConverter.from_decision_proto(
          suggest_decision_proto
      )

      # Write the metadata update to the datastore.
      try:
        self.datastore.update_metadata(
            study_name,
            svz.metadata_util.make_key_value_list(
                suggest_decision.metadata.on_study
            ),
            svz.metadata_util.trial_metadata_to_update_list(
                suggest_decision.metadata.on_trials
            ),
        )
      except KeyError as e:
        output_op.error.CopyFrom(
            status_pb2.Status(code=code_pb2.Code.INTERNAL, message=str(e))
        )
        logging.exception(
            'Failed to write metadata update to datastore: %s',
            suggest_decision.metadata,
        )
        output_op.done = True
        self.datastore.update_suggestion_operation(output_op)
        return output_op

      new_py_trials = [
          decision.to_trial() for decision in suggest_decision.suggestions
      ]
      new_trials = svz.TrialConverter.to_protos(new_py_trials)

      while request.suggestion_count > len(output_trials):
        new_trial = new_trials.pop()
        trial_id = self.datastore.max_trial_id(request.parent) + 1
        new_trial.id = str(trial_id)
        new_trial.name = TrialResource(owner_id, study_id, trial_id).name
        new_trial.state = study_pb2.Trial.State.ACTIVE
        new_trial.start_time.CopyFrom(start_time)
        new_trial.client_id = request.client_id
        self.datastore.create_trial(new_trial)
        output_trials.append(new_trial)

      output_op.response.value = vizier_service_pb2.SuggestTrialsResponse(
          trials=output_trials, start_time=start_time
      ).SerializeToString()

      # Store remaining trials as REQUESTED if Pythia over-delivered.
      for remain_trial in new_trials:
        trial_id = self.datastore.max_trial_id(request.parent) + 1
        remain_trial.id = str(trial_id)
        remain_trial.name = TrialResource(owner_id, study_id, trial_id).name
        remain_trial.state = study_pb2.Trial.State.REQUESTED
        self.datastore.create_trial(remain_trial)

      output_op.done = True
      self.datastore.update_suggestion_operation(output_op)
      return output_op

  def GetOperation(
      self,
      request: operations_pb2.GetOperationRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> operations_pb2.Operation:
    """Gets the latest state of a SuggestTrials() long-running operation."""
    return self.datastore.get_suggestion_operation(request.name)

  def CreateTrial(
      self,
      request: vizier_service_pb2.CreateTrialRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Trial:
    """Adds user provided Trial to a Study and assigns the correct fields."""
    if self._study_is_immutable(request.parent):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot create trial.'.format(request.parent)
      )
      grpc_util.handle_exception(e, context)

    trial = request.trial
    with self._study_name_to_lock[request.parent]:
      trial.id = str(self.datastore.max_trial_id(request.parent) + 1)
      study_resource = StudyResource.from_name(request.parent)
      trial.name = (study_resource.trial_resource(trial.id)).name

      if trial.state != study_pb2.Trial.State.SUCCEEDED:
        trial.state = study_pb2.Trial.State.REQUESTED
      trial.ClearField('client_id')

      trial.start_time.CopyFrom(_get_current_time())
      self.datastore.create_trial(trial)
    return trial

  def GetTrial(
      self,
      request: vizier_service_pb2.GetTrialRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> Optional[study_pb2.Trial]:
    """Gets a Trial."""
    return self.datastore.get_trial(request.name)

  def ListTrials(
      self,
      request: vizier_service_pb2.ListTrialsRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> vizier_service_pb2.ListTrialsResponse:
    """Lists the Trials associated with a Study."""
    list_of_trials = self.datastore.list_trials(request.parent)
    return vizier_service_pb2.ListTrialsResponse(trials=list_of_trials)

  def AddTrialMeasurement(
      self,
      request: vizier_service_pb2.AddTrialMeasurementRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Trial:
    """Adds a measurement of the objective metrics to a Trial.

    This measurement is assumed to have been taken before the Trial is
    complete.

    Args:
      request:
      context:

    Returns:
      Trial whose measurement was appended.

    Raises:
      ImmutableStudyError: If study was already immutable.
      ImmutableTrialError: If the trial cannot be modified.
    """
    study_name = TrialResource.from_name(request.trial_name).study_resource.name
    if self._study_is_immutable(study_name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot add measurement.'.format(study_name)
      )
      grpc_util.handle_exception(e, context)

    with self._study_name_to_lock[study_name]:
      trial = self.datastore.get_trial(request.trial_name)
      if trial.state == study_pb2.Trial.State.INFEASIBLE:
        return trial
      if trial.state not in self._TRIAL_MUTABLE_STATES:
        e = custom_errors.ImmutableTrialError(
            'Trial {} has state {}. Measurements can only be added to trials in'
            ' state ACTIVE or STOPPING'.format(
                request.trial_name, study_pb2.Trial.State.Name(trial.state)
            )
        )
        grpc_util.handle_exception(e, context)

      trial.measurements.extend([request.measurement])
      self.datastore.update_trial(trial)
    return trial

  # TODO: Auto selection defaults to the last measurement.
  # Add support for "best measurement" behavior.
  def CompleteTrial(
      self,
      request: vizier_service_pb2.CompleteTrialRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Trial:
    """Marks a Trial as complete."""
    study_name = TrialResource.from_name(request.name).study_resource.name
    if self._study_is_immutable(study_name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot complete trial.'.format(study_name)
      )
      grpc_util.handle_exception(e, context)

    with self._study_name_to_lock[study_name]:
      trial = self.datastore.get_trial(request.name)
      if trial.state not in self._TRIAL_MUTABLE_STATES:
        e = custom_errors.ImmutableTrialError(
            'Trial {} has state {}. Only trials in state ACTIVE or STOPPING '
            'can be completed.'.format(
                request.name, study_pb2.Trial.State.Name(trial.state)
            )
        )
        grpc_util.handle_exception(e, context)

      trial.state = study_pb2.Trial.State.SUCCEEDED
      if request.final_measurement.metrics:
        trial.final_measurement.CopyFrom(request.final_measurement)
      elif not request.trial_infeasible:
        # Trial's final measurement auto-selected from latest reported
        # measurement.
        if not trial.measurements:
          e = ValueError(
              'Both the request and trial intermediate measurements are'
              " missing. Cannot determine trial's final_measurement."
          )
          grpc_util.handle_exception(e, context)
        trial.final_measurement.CopyFrom(trial.measurements[-1])

      # Handle infeasibility.
      if request.trial_infeasible:
        trial.state = study_pb2.Trial.State.INFEASIBLE
        trial.infeasible_reason = request.infeasible_reason

      self.datastore.update_trial(trial)
    return trial

  def DeleteTrial(
      self,
      request: vizier_service_pb2.DeleteTrialRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> empty_pb2.Empty:
    """Deletes a Trial."""
    study_name = TrialResource.from_name(request.name).study_resource.name
    if self._study_is_immutable(study_name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot delete trial.'.format(study_name)
      )
      grpc_util.handle_exception(e, context)

    self.datastore.delete_trial(request.name)
    return empty_pb2.Empty()

  # TODO: This currently uses the same algorithm as suggestion.
  def CheckTrialEarlyStoppingState(
      self,
      request: vizier_service_pb2.CheckTrialEarlyStoppingStateRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> vizier_service_pb2.CheckTrialEarlyStoppingStateResponse:
    """Checks whether a Trial should stop or not.

    The logic is as follows:

    1. We check if there already exists an operation. If not, create it. But if
    it already exists:
      a. If it's still ACTIVE (i.e. early stopping is still being computed by
      Pythia), just return it.
      b. If it's done and the call is very recent, just return it.
      c. If it's been a long time, we can recycle the operation, and remark it
      as ACTIVE to be used again.

    2. Send the operation to Pythia, which will return a set of early stopping
    decisions. The Pythia policy is guaranteed to return a decision for the
    requested trial. However, Pythia policy may also signal whether other trials
    should stop or not (e.g. in algorithms in which proposed trials are batched
    and correlated).

    3. Transfer these decisions' `should_stop` into the relevant operations
    (some of which will also be created in this call, for future use), including
    the original operation which needs to be returned. Mark them all as done.

    4. Return the original trial's `should_stop` decision.

    Args:
      request:
      context:

    Returns:
      A CheckTrialEarlyStoppingStateResponse, containing a bool to denote
      whether the requested trial should stop.

    Raises:
      ImmutableStudyError: If study was already immutable.
      ImmutableTrialError: If the trial cannot be modified.
    """
    trial_resource = TrialResource.from_name(request.trial_name)
    study_name = trial_resource.study_resource.name
    if self._study_is_immutable(study_name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot early stop trial.'.format(study_name)
      )
      grpc_util.handle_exception(e, context)

    with self._study_name_to_lock[study_name]:
      trial = self.datastore.get_trial(request.trial_name)
      if trial.state not in self._TRIAL_MUTABLE_STATES:
        e = custom_errors.ImmutableTrialError(
            'Trial {} has state {}. Only trials in state ACTIVE or STOPPING '
            'can be completed.'.format(
                request.trial_name, study_pb2.Trial.State.Name(trial.state)
            )
        )
        grpc_util.handle_exception(e, context)
    outer_op_name = trial_resource.early_stopping_operation_resource.name
    # Don't allow simultaneous SuggestTrial or EarlyStopping calls to be
    # processed.
    with self._operation_lock[study_name]:
      try:
        # Reuse any existing early stopping op, since the Pythia policy may have
        # already signaled this trial to stop.
        output_operation = self.datastore.get_early_stopping_operation(
            outer_op_name
        )
      except KeyError:
        output_operation = None

      if output_operation is None:
        # Create fresh new operation.
        output_operation = vizier_oss_pb2.EarlyStoppingOperation(
            name=outer_op_name,
            status=vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE,
            should_stop=False,
        )
        output_operation.creation_time.CopyFrom(_get_current_time())
        self.datastore.create_early_stopping_operation(output_operation)
      else:
        if (
            output_operation.status
            == vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE
            or datetime.datetime.utcnow()
            - output_operation.completion_time.ToDatetime()
            < self._early_stop_recycle_period
        ):
          # Operation is already active or very recent. Just return it.
          return vizier_service_pb2.CheckTrialEarlyStoppingStateResponse(
              should_stop=output_operation.should_stop
          )

        # Recycle the operation to ACTIVE again and start Pythia for
        # recomputation.
        output_operation.status = (
            vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE
        )
        output_operation.should_stop = False  # Defaulted back to False.
        self.datastore.update_early_stopping_operation(output_operation)

      study = self.datastore.load_study(study_name)
      study_config = svz.StudyConfig.from_proto(study.study_spec)
      study_descriptor = vz.StudyDescriptor(
          config=study_config,
          guid=study_name,
          max_trial_id=self.datastore.max_trial_id(study_name),
      )
      early_stop_request = pythia.EarlyStopRequest(
          study_descriptor=study_descriptor, trial_ids=[trial_resource.trial_id]
      )
      early_stop_request_proto = svz.EarlyStopConverter.to_request_proto(
          early_stop_request
      )
      spec_name = (
          study.study_spec.WhichOneof('automated_stopping_spec')
          or 'default_stopping_spec'
      )
      if spec_name == 'default_stopping_spec':
        # TODO: Add TSGP algorithm when open sourced.
        early_stop_request_proto.algorithm = 'RANDOM_SEARCH'
      else:
        raise ValueError(
            f'Misconfigured automated_stopping_spec: {study.study_spec}'
        )

      # Send request to Pythia.
      temp_pythia_service = self._select_pythia_service(
          study_config.pythia_endpoint
      )
      early_stopping_decisions_proto = temp_pythia_service.EarlyStop(
          early_stop_request_proto
      )
      early_stopping_decisions = svz.EarlyStopConverter.from_decisions_proto(
          early_stopping_decisions_proto
      )
      # Update metadata from result.
      self.datastore.update_metadata(
          study_name,
          svz.metadata_util.make_key_value_list(
              early_stopping_decisions.metadata.on_study
          ),
          svz.metadata_util.trial_metadata_to_update_list(
              early_stopping_decisions.metadata.on_trials
          ),
      )

      # Pythia does not guarantee that the output_operation's id
      # will be in the decisions.
      for early_stopping_decision in early_stopping_decisions.decisions:
        inner_op_name = resources.EarlyStoppingOperationResource(
            trial_resource.owner_id,
            trial_resource.study_id,
            early_stopping_decision.id,
        ).name
        try:
          inner_operation = self.datastore.get_early_stopping_operation(
              inner_op_name
          )
        except KeyError:
          # Create the operation to store early stopping data for future use.
          inner_operation = vizier_oss_pb2.EarlyStoppingOperation(
              name=inner_op_name,
              status=vizier_oss_pb2.EarlyStoppingOperation.Status.ACTIVE,
              should_stop=False,
          )
          inner_operation.creation_time.CopyFrom(_get_current_time())
          self.datastore.create_early_stopping_operation(inner_operation)

        inner_operation.should_stop = early_stopping_decision.should_stop
        inner_operation.status = (
            vizier_oss_pb2.EarlyStoppingOperation.Status.DONE
        )
        inner_operation.completion_time.CopyFrom(_get_current_time())
        self.datastore.update_early_stopping_operation(inner_operation)

      # Operation to be outputted may have changed.
      output_operation = self.datastore.get_early_stopping_operation(
          output_operation.name
      )
      return vizier_service_pb2.CheckTrialEarlyStoppingStateResponse(
          should_stop=output_operation.should_stop
      )

  def StopTrial(
      self,
      request: vizier_service_pb2.StopTrialRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> study_pb2.Trial:
    """Sets the trial state to STOPPING.

    Args:
      request:
      context:

    Returns:
      The stopped Trial

    Raises:
      ImmutableStudyError: If study was already immutable.
      ImmutableTrialError: If the trial cannot be modified.
    """
    study_name = TrialResource.from_name(request.name).study_resource.name
    if self._study_is_immutable(study_name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot stop trial.'.format(study_name)
      )
      grpc_util.handle_exception(e, context)

    with self._study_name_to_lock[study_name]:
      trial = self.datastore.get_trial(request.name)
      if trial.state == study_pb2.Trial.ACTIVE:
        trial.state = study_pb2.Trial.STOPPING
        self.datastore.update_trial(trial)
      elif trial.state in (study_pb2.Trial.STOPPING, study_pb2.Trial.SUCCEEDED):
        logging.warn(
            'Trial %s has state %s. StopTrial will be a no-op.',
            request.name,
            study_pb2.Trial.State.Name(trial.state),
        )
      else:
        e = custom_errors.ImmutableTrialError(
            'Trial {} has state {}. Raising Error.'.format(
                request.name, study_pb2.Trial.State.Name(trial.state)
            )
        )
        grpc_util.handle_exception(e, context)
    return trial

  def ListOptimalTrials(
      self,
      request: vizier_service_pb2.ListOptimalTrialsRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> vizier_service_pb2.ListOptimalTrialsResponse:
    """The definition of pareto-optimal can be checked in wiki page.

    https://en.wikipedia.org/wiki/Pareto_efficiency.

    Args:
      request:
      context:

    Returns:
      A list containing pareto-optimal Trials for multi-objective Study or the
      optimal Trials for single-objective Study.
    """
    raw_trial_list = self.datastore.list_trials(request.parent)
    if not raw_trial_list:
      return vizier_service_pb2.ListOptimalTrialsResponse(optimal_trials=[])

    study_spec = self.datastore.load_study(request.parent).study_spec
    metric_id_to_goal = {m.metric_id: m.goal for m in study_spec.metrics}
    required_metric_ids = set(metric_id_to_goal.keys())

    considered_trials = []
    considered_trial_objective_vectors = []

    for trial in raw_trial_list:
      trial_metric_id_to_value = {
          m.metric_id: m.value for m in trial.final_measurement.metrics
      }
      trial_metric_ids = set(trial_metric_id_to_value.keys())
      # Add trials ONLY if they succeeded and contain all supposed metrics.
      if (
          trial.state == study_pb2.Trial.State.SUCCEEDED
          and required_metric_ids.issubset(trial_metric_ids)
      ):
        objective_vector = []
        for metric_id, goal in metric_id_to_goal.items():
          # Flip sign for convenience when computing optimality.
          if goal == study_pb2.StudySpec.MetricSpec.GoalType.MINIMIZE:
            vector_value = -1.0 * trial_metric_id_to_value[metric_id]
          else:
            vector_value = trial_metric_id_to_value[metric_id]
          objective_vector.append(vector_value)

        considered_trials.append(trial)
        considered_trial_objective_vectors.append(objective_vector)

    if not considered_trials:
      return vizier_service_pb2.ListOptimalTrialsResponse(optimal_trials=[])

    # Find Pareto optimal trials.
    ys = np.array(considered_trial_objective_vectors)
    n = ys.shape[0]
    dominated = np.asarray([
        [np.all(ys[i] <= ys[j]) & np.any(ys[j] > ys[i]) for i in range(n)]
        for j in range(n)
    ])
    optimal_booleans = np.logical_not(np.any(dominated, axis=0))
    optimal_trials = []
    for i, boolean in enumerate(list(optimal_booleans)):
      if boolean:
        optimal_trials.append(considered_trials[i])

    return vizier_service_pb2.ListOptimalTrialsResponse(
        optimal_trials=optimal_trials
    )

  def UpdateMetadata(
      self,
      request: vizier_service_pb2.UpdateMetadataRequest,
      context: Optional[grpc.ServicerContext] = None,
  ) -> vizier_service_pb2.UpdateMetadataResponse:
    """Stores the supplied metadata in the database."""
    if self._study_is_immutable(request.name):
      e = custom_errors.ImmutableStudyError(
          'Study {} is immutable. Cannot update metadata.'.format(request.name)
      )
      grpc_util.handle_exception(e, context)

    try:
      self.datastore.update_metadata(
          request.name,
          [x.metadatum for x in request.delta if not x.HasField('trial_id')],
          [x for x in request.delta if x.HasField('trial_id')],
      )
    except KeyError as e:
      return vizier_service_pb2.UpdateMetadataResponse(
          error_details=';'.join(e.args)
      )
    return vizier_service_pb2.UpdateMetadataResponse()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Thin wrappers around Jax optimizers."""
# pylint: disable=unused-import

import functools

from vizier._src.jax.optimizers.core import LossFunction
from vizier._src.jax.optimizers.core import Optimizer
from vizier._src.jax.optimizers.core import Params
from vizier._src.jax.optimizers.jaxopt_wrappers import JaxoptLbfgsB
from vizier._src.jax.optimizers.jaxopt_wrappers import JaxoptScipyLbfgsB
from vizier._src.jax.optimizers.jaxopt_wrappers import LbfgsBOptions
from vizier._src.jax.optimizers.optax_wrappers import OptaxTrain

DEFAULT_RANDOM_RESTARTS = 4


@functools.lru_cache
def default_optimizer(maxiter: int = 50) -> Optimizer:
  """Default optimizer and random restarts that work okay for most cases."""
  # NOTE: Production algorithms are recommended to stay away from using this.
  return JaxoptScipyLbfgsB(LbfgsBOptions(maxiter=maxiter, best_n=None))



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Utils functions to support Eagle Strategy designer."""

import collections
import copy
import math
from typing import DefaultDict, Dict, Optional
from absl import logging
import attr
import numpy as np
from vizier import pyvizier as vz
from vizier._src.algorithms.random import random_sample

# Standardize name used to assign for trials' metric name, which simplifies
# serializing trials as part of the FireflyPool.
OBJECTIVE_NAME = 'objective'


@attr.define
class FireflyAlgorithmConfig:
  """Configuration hyperparameters for Eagle Strategy / Firefly Algorithm."""
  # ---------------------------- Gravity ---------------------------------------
  # The amount of pulling force a brighter firefly applies to its neighbors.
  gravity: float = 1.0
  # The amount of repel force a darker firefly applies to its neighbors.
  negative_gravity: float = 0.02

  # --------------------------- Visibility -------------------------------------
  # The visibility associated with continuous parameters.
  visibility: float = 3.0
  # The visibility associated with categorical parameters.
  categorical_visibility: float = 0.2
  # The visibility associated with discrete parameters.
  discrete_visibility: float = 1.0

  # ---------------------------- Perturbation ----------------------------------
  # The default perturbation that fireflies start with.
  perturbation: float = 1e-1
  # The lower bound of the perturbation factor below which firefly are removed.
  perturbation_lower_bound: float = 1e-3
  # The perturbation factor for categorical parameters.
  categorical_perturbation_factor: float = 25.0
  # The perturbation factor for discrete parameters.
  discrete_perturbation_factor: float = 10.0
  # The perturbation of purely categorical search space.
  pure_categorical_perturbation: float = 0.1
  # The maximum perturbation factor.
  max_perturbation: float = 0.5
  # The penalize factor used to penalize a parent firefly's lack of improvement.
  penalize_factor: float = 0.9

  # ---------------------------- Exploration -----------------------------------
  # Exploration rate (value > 1.0 encourages more exploration)
  explore_rate: float = 1.0
  # The factor to apply on infeasible trial repel force.
  infeasible_force_factor: float = 0.0

  # ----------------------------- Pool Size ------------------------------------
  # The factor used to compute the pool size.
  pool_size_factor: float = 1.2
  # The maximum pool size.
  max_pool_size: int = 1000


@attr.define
class Firefly:
  """The Firefly class represents a single firefly in the pool.

  Attributes:
    id_: A unique firefly identifier. This is used to associate trials with
      their parent fireflies.
    perturbation: Controls the amount of exploration. Signifies the amount of
      perturbation to add to the generated trial parameters. The value of
      'perturbation' keeps decreasing if the suggested trial doesn't improve the
      objective function until it reaches 'perturbation_lower_bound' in which
      case we remove the firefly from the pool.
    generation: The number of "successful" (better than the last trial) trials
      suggested from the firefly.
    trial: The best trial associated with the firefly.
  """
  id_: int
  perturbation: float
  generation: int
  trial: vz.Trial


@attr.define
class EagleStrategyUtils:
  """Eagle Strategy utils.

  Attributes:
    search_space: The search space that Eagle attempts to optimize on.
    config: The Eagle Strategy configuration.
    rng: The random generator to sample random distributions.
  """
  problem_statement: vz.ProblemStatement
  config: FireflyAlgorithmConfig
  rng: np.random.Generator
  _search_space: vz.SearchSpace = attr.field(init=False)
  _n_parameters: int = attr.field(init=False)
  _degrees_of_freedom: DefaultDict[vz.ParameterType, int] = attr.field(
      init=False, factory=lambda: collections.defaultdict(int))
  _original_metric_name: str = attr.field(init=False)
  _goal: vz.ObjectiveMetricGoal = attr.field(init=False)

  def __attrs_post_init__(self):
    """Initialize and cache common values and objects."""
    self._search_space = self.problem_statement.search_space
    self._n_parameters = len(self._search_space.parameters)
    self._cache_degrees_of_freedom()
    self._original_metric_name = (
        self.problem_statement.single_objective_metric_name
    )
    self._goal = self.problem_statement.metric_information.item().goal
    logging.info('EagleStrategyUtils instance was created.\n%s', str(self))

  def compute_pull_weight_by_type(
      self,
      other_parameters: vz.ParameterDict,
      suggested_parameters: vz.ParameterDict,
      is_other_fly_better: bool,
  ) -> Dict[vz.ParameterType, float]:
    """Computes the pull weights by type."""
    # Compute squared distances between the vector of parameters of each type.
    squared_distances = self._compute_canonical_distance_squared_by_type(
        other_parameters, suggested_parameters)
    # Determine the direction (attraction vs. repulsion).
    if is_other_fly_better > 0:
      pull_direction = self.config.gravity
    else:
      pull_direction = -self.config.negative_gravity

    pull_weights = {}
    # Iterate over the squared distance by type and compute the pull force.
    for param_type, squared_distance in squared_distances.items():
      degree_of_freedom = self._degrees_of_freedom[param_type]
      if degree_of_freedom == 0:
        pull_weights[param_type] = 0
      else:
        scaled_squared_distance = squared_distance / degree_of_freedom * 10
        # Determine the visibilty based on the parameter type.
        if param_type == vz.ParameterType.CATEGORICAL:
          visiblity = self.config.categorical_visibility
        elif param_type in [
            vz.ParameterType.DISCRETE, vz.ParameterType.INTEGER
        ]:
          visiblity = self.config.discrete_visibility
        elif param_type == vz.ParameterType.DOUBLE:
          visiblity = self.config.visibility
        else:
          raise ValueError('Unsupported parameter type: %s' % param_type)
        # Compute the pull weight and insert to dictionary.
        pull_weights[param_type] = math.exp(
            -visiblity * scaled_squared_distance) * pull_direction

    return pull_weights

  @property
  def _param_perturb_scales(self):
    """Set the parameter perturbation scaling."""
    scales = [1.0 for _ in range(self._n_parameters)]
    for i, param_config in enumerate(self._search_space.parameters):
      if param_config.type == vz.ParameterType.CATEGORICAL:
        # For CATEGORICAL parameters, the perturbation is interpreted as a
        # probability of replacing the value with a uniformly random category.
        scales[i] = self.config.categorical_perturbation_factor
      if param_config.type == vz.ParameterType.DISCRETE:
        scales[i] = self.config.discrete_perturbation_factor / (
            param_config.num_feasible_values * self.config.perturbation)
    return np.asarray(scales)

  def _cache_degrees_of_freedom(self) -> None:
    """Computes the degrees of freedom for each parameter type and cache it.

    The function counts the number of parameter configs in the search space that
    has more than only a single feasible point.
    """
    for param_config in self._search_space.parameters:
      if param_config.num_feasible_values > 1:
        self._degrees_of_freedom[param_config.type] += 1

  def _compute_canonical_distance_squared_by_type(
      self,
      p1: vz.ParameterDict,
      p2: vz.ParameterDict,
  ) -> Dict[vz.ParameterType, float]:
    """Computes the canonical distance squared by parameter type."""
    dist_squared_by_type = {
        vz.ParameterType.DOUBLE: 0.0,
        vz.ParameterType.DISCRETE: 0.0,
        vz.ParameterType.INTEGER: 0.0,
        vz.ParameterType.CATEGORICAL: 0.0
    }

    for param_config in self._search_space.parameters:
      p1_value = p1[param_config.name].value
      p2_value = p2[param_config.name].value

      if param_config.type == vz.ParameterType.CATEGORICAL:
        dist_squared_by_type[param_config.type] += int(p1_value == p2_value)
      else:
        min_value, max_value = param_config.bounds
        if max_value == min_value:
          dist = 0
        else:
          dist = (p1_value - p2_value) / (max_value - min_value)
        dist_squared_by_type[param_config.type] += dist * dist

    return dist_squared_by_type

  def compute_cononical_distance(
      self,
      p1: vz.ParameterDict,
      p2: vz.ParameterDict,
  ) -> float:
    """Computes the canonical squared distance between two parameters."""
    dist_by_type = self._compute_canonical_distance_squared_by_type(p1, p2)
    return sum(dist_by_type.values())

  def compute_pool_capacity(self) -> int:
    """Computes the pool capacity."""
    df = self._n_parameters
    return min(
        10 + round((df**self.config.pool_size_factor + df) * 0.5),
        self.config.max_pool_size,
    )

  def combine_two_parameters(
      self,
      param_config: vz.ParameterConfig,
      param1: vz.ParameterDict,
      param2: vz.ParameterDict,
      param1_weight: float,
  ) -> vz.ParameterValueTypes:
    """Combnies the values of two parameters based on their weights and type.

    For decimal parameters, performs a linear combination of two parameters,
    by computing f1 * f1_weight + f2 * (1 - f1_weight). Note that this is not a
    convex combination, because 'f1_weight' can be outside [0, 1].

    For categorical parameters, uses Bernuolli to choose between the value of
    the first parameter and the second with probability that equals the weight.

    Args:
      param_config:
      param1:
      param2:
      param1_weight:

    Returns:
      The combined weighted value of the two parameters.
    """
    value1 = param1[param_config.name].value
    value2 = param2[param_config.name].value
    if param_config.type == vz.ParameterType.CATEGORICAL:
      if 0.0 < param1_weight < 1.0:
        prob1 = param1_weight
        new_value = random_sample.sample_bernoulli(
            self.rng, prob1, value1, value2
        )
      elif param1_weight <= 0.0:
        new_value = value2
      else:  # param1_weight >= 1.0
        new_value = value1
    else:
      weighted_param_value = value1 * param1_weight + value2 * (
          1 - param1_weight
      )
      if param_config.type == vz.ParameterType.DOUBLE:
        new_value = weighted_param_value
      elif param_config.type == vz.ParameterType.INTEGER:
        new_value = round(weighted_param_value)
      elif param_config.type == vz.ParameterType.DISCRETE:
        new_value = random_sample.get_closest_element(
            param_config.feasible_values, weighted_param_value)
      else:
        raise ValueError('Invalid parameter type: %s' % param_config.type)
      new_value = min(new_value, param_config.bounds[1])
      new_value = max(new_value, param_config.bounds[0])
    return new_value

  def create_perturbations(self, perturbation: float) -> list[float]:
    """Creates perturbations array."""
    if self.is_pure_categorical():
      return [
          self.config.pure_categorical_perturbation
          for _ in range(self._n_parameters)
      ]
    perturbations = self.rng.laplace(size=(self._n_parameters,))
    perturbation_direction = perturbations / max(abs(perturbations))
    perturbations = perturbation_direction * perturbation
    return [float(x) for x in perturbations * self._param_perturb_scales]

  def perturb_parameter(
      self,
      param_config: vz.ParameterConfig,
      value: vz.ParameterValueTypes,
      perturbation: float,
  ) -> vz.ParameterValueTypes:
    """Perturbs the parameter based on its type and the amount of perturbation.

    For categorical parameters, with probability 'perturbation', replace with a
    uniformly random category.

    For numeric parameters, 'perturbation' determines the fraction of the
    parameter range to add, where perturbation=-1 will always return the minimum
    possible value, and perturbation=+1 will return the maximum possible value.

    Args:
      param_config: The configuration of the parameter
      value: The parameter value before perturbation
      perturbation: The amount of pertrubation to apply

    Returns:
      The parameter value after perturbation.

    Raises:
      Exception: if the parameter has invalid type.
    """
    if param_config.type == vz.ParameterType.CATEGORICAL:
      if random_sample.sample_uniform(self.rng) < abs(perturbation):
        return random_sample.sample_categorical(self.rng,
                                                param_config.feasible_values)
      else:
        return value

    min_value, max_value = param_config.bounds
    perturb_val = value + perturbation * (max_value - min_value)

    if param_config.type == vz.ParameterType.DISCRETE:
      return random_sample.get_closest_element(param_config.feasible_values,
                                               perturb_val)

    perturb_val = min(perturb_val, param_config.bounds[1])
    perturb_val = max(perturb_val, param_config.bounds[0])
    if param_config.type == vz.ParameterType.DOUBLE:
      return perturb_val
    elif param_config.type == vz.ParameterType.INTEGER:
      return round(perturb_val)
    else:
      raise ValueError('Invalid parameter type: %s' % param_config.type)

  def get_metric(self, trial: vz.Trial) -> float:
    """Returns the trial metric."""
    if trial.infeasible:
      return np.nan
    return trial.final_measurement.metrics[OBJECTIVE_NAME]  # pytype: disable=bad-return-type

  def is_better_than(
      self,
      trial1: vz.Trial,
      trial2: vz.Trial,
  ) -> bool:
    """Checks whether the 'trial1' is better than 'trial2'.

    The comparison is based on the value of final measurement and whether it
    goal is MAXIMIZATION or MINIMIZATON.

    If either trials is not completed, infeasible or missing the final
    measurement returns False.

    Args:
      trial1:
      trial2:

    Returns:
      Whether trial1 is greater than trial2.
    """
    if not trial1.is_completed or not trial2.is_completed:
      return False
    if trial1.infeasible and not trial2.infeasible:
      return False
    if not trial1.infeasible and trial2.infeasible:
      return True
    if trial1.final_measurement is None or trial2.final_measurement is None:
      return False

    if self._goal == vz.ObjectiveMetricGoal.MAXIMIZE:
      return trial1.final_measurement.metrics[
          OBJECTIVE_NAME] > trial2.final_measurement.metrics[OBJECTIVE_NAME]
    else:
      return trial1.final_measurement.metrics[
          OBJECTIVE_NAME] < trial2.final_measurement.metrics[OBJECTIVE_NAME]

  def is_pure_categorical(self) -> bool:
    """Returns True if all parameters in search_space are categorical."""
    return all([
        p.type == vz.ParameterType.CATEGORICAL
        for p in self._search_space.parameters
    ])

  def standardize_trial_metric_name(self, trial: vz.Trial) -> vz.Trial:
    """Creates a new trial with canonical metric name."""
    if trial.infeasible:
      return trial
    value = trial.final_measurement.metrics[self._original_metric_name].value
    new_trial = vz.Trial(parameters=trial.parameters, metadata=trial.metadata)
    new_trial.complete(
        measurement=vz.Measurement(metrics={OBJECTIVE_NAME: value}))
    return new_trial

  def display_trial(self, trial: vz.Trial) -> str:
    """Construct a string to represent a completed trial."""
    parameters = {
        k: v if isinstance(v, str) else round(v, 3)
        for k, v in trial.parameters.as_dict().items()
    }
    if trial.final_measurement:
      obj_value = (
          f'{list(trial.final_measurement.metrics.values())[0].value:.5f}'
      )
      return f'Value: {obj_value}, Parameters: {parameters}'
    else:
      return f'Parameters: {parameters}'


@attr.define
class FireflyPool:
  """The class maintains the Firefly pool and relevent operations.

  Attributes:
    utils: Eagle Strategy utils class.
    capacity: The maximum number of non-feasible fireflies in the pool.
    size: The current number of non-feasible fireflies in the pool.
    _pool: A dictionary of Firefly objects organized by firefly id.
    _last_id: The last firefly id used to generate a suggestion. It's persistent
      across calls to ensure we don't use the same fly repeatedly.
    _max_fly_id: The maximum value of any fly id ever created. It's persistent
      persistent accross calls to ensure unique ids even if trails were deleted.
    _infeasible_count: The number of infeasible fireflies in the pool.
  """
  _utils: EagleStrategyUtils
  _capacity: int
  _pool: Dict[int, Firefly] = attr.field(init=False, default=attr.Factory(dict))
  _last_id: int = attr.field(init=False, default=0)
  _max_fly_id: int = attr.field(init=False, default=0)
  _infeasible_count: int = attr.field(init=False, default=0)

  @property
  def capacity(self) -> int:
    return self._capacity

  @property
  def size(self) -> int:
    """Returns the number of feasible fireflies in the pool."""
    return len(self._pool) - self._infeasible_count

  def remove_fly(self, fly: Firefly):
    """Removes a fly from the pool."""
    if fly.trial.infeasible:
      raise ValueError('Infeasible firefly should not be removed from pool.')
    del self._pool[fly.id_]

  def get_shuffled_flies(self, rng: np.random.Generator) -> list[Firefly]:
    """Shuffles the fireflies and returns them as a list."""
    return random_sample.shuffle_list(rng, list(self._pool.values()))

  def generate_new_fly_id(self) -> int:
    """Generates a unique fly id (starts from 0) to identify a fly in the pool."""
    self._max_fly_id += 1
    return self._max_fly_id - 1

  def get_next_moving_fly_copy(self) -> Firefly:
    """Finds the next fly, returns a copy of it and updates '_last_id'.

    To find the next moving fly, we start from index of '_last_id'+1 and
    incremently check whether the index exists in the pool. When the loop
    reaches 'max_fly_id' it goes back to index 0. We return a copy of the
    first fly we find an index for. Before returning we also set '_last_id'
    to the next moving fly id.

    Note that we don't assume the existance of '_last_id' in the pool, as the
    fly with `_last_id` might be removed from the pool.

    Returns:
      A copy of the next moving fly.
    """
    curr_id = self._last_id + 1
    while curr_id != self._last_id:
      if curr_id > self._max_fly_id:
        # Passed the maximum id. Start from the first one as ids are monotonic.
        curr_id = next(iter(self._pool))
      if curr_id in self._pool and not self._pool[curr_id].trial.infeasible:
        self._last_id = curr_id
        return copy.deepcopy(self._pool[curr_id])
      curr_id += 1

    return copy.deepcopy(self._pool[self._last_id])

  def is_best_fly(self, fly: Firefly) -> bool:
    """Checks if the 'fly' has the best final measurement in the pool."""
    for other_fly_id, other_fly in self._pool.items():
      if other_fly_id != fly.id_ and self._utils.is_better_than(
          other_fly.trial, fly.trial
      ):
        return False
    return True

  def find_parent_fly(self, parent_fly_id: Optional[int]) -> Optional[Firefly]:
    """Obtains the parent firefly associated with the trial.

    Extract the associated parent id from the trial's metadata and attempt
    to find the parent firefly in the pool. If it doesn't exist return None.

    Args:
      parent_fly_id:

    Returns:
      Firefly or None.
    """
    parent_fly = self._pool.get(parent_fly_id, None)
    return parent_fly

  def find_closest_parent(self, trial: vz.Trial) -> Firefly:
    """Finds the closest fly in the pool to a given trial."""
    if not self._pool:
      raise ValueError('Pool was empty when searching for closest parent.')

    min_dist, closest_parent = float('inf'), next(iter(self._pool.values()))
    for other_fly in self._pool.values():
      if other_fly.trial.infeasible:
        continue
      curr_dist = self._utils.compute_cononical_distance(
          other_fly.trial.parameters, trial.parameters
      )
      if curr_dist < min_dist:
        min_dist = curr_dist
        closest_parent = other_fly

    return closest_parent

  def create_or_update_fly(self, trial: vz.Trial, parent_fly_id: int) -> None:
    """Creates a new fly in the pool or update an existing one.

    This method is called when the pool is below capacity.

    The newly created fly is assigned with 'id_' equals 'parent_fly_id' which
    is taken from the trial's metadata. The fly's id is determined during the
    'Suggest' method and stored in the metadata.

    Edge case: if the 'parent_fly_id' already exists in the pool (and the pool
    is below capacity) we update the matching fly with the better trial. This
    scenario could happen if for example a batch larger than the pool capacity
    was suggested, and then trials associated with the same fly were reported
    as COMPLETED sooner than the other trials, so when the pool is updated
    during 'Update' we encounter trials from the same fly before the pool is at
    capactiy.

    Args:
      trial:
      parent_fly_id:
    """
    if parent_fly_id not in self._pool:
      # Create a new Firefly in pool.
      new_fly = Firefly(
          id_=parent_fly_id,
          perturbation=self._utils.config.perturbation,
          generation=1,
          trial=trial,
      )
      self._pool[parent_fly_id] = new_fly
      if trial.infeasible:
        self._infeasible_count += 1
    else:
      # Parent fly id already in pool. Update trial if there was improvement.
      if self._utils.is_better_than(trial, self._pool[parent_fly_id].trial):
        self._pool[parent_fly_id].trial = trial



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Random vectorized optimizer to be used in convergence test."""

from typing import Optional

import jax
from jax import numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp
from vizier._src.algorithms.optimizers import vectorized_base as vb
from vizier._src.jax import types
from vizier.pyvizier import converters

tfd = tfp.distributions


class RandomVectorizedStrategy(vb.VectorizedStrategy[None]):
  """Random vectorized strategy."""

  def __init__(
      self,
      converter: converters.TrialToModelInputConverter,
      suggestion_batch_size: int,
  ):
    empty_features = converter.to_features([])
    n_feature_dimensions_with_padding = types.ContinuousAndCategorical(
        empty_features.continuous.shape[-1],
        empty_features.categorical.shape[-1],
    )

    categorical_sizes = []
    for spec in converter.output_specs.categorical:
      categorical_sizes.append(spec.bounds[1])

    self._suggestion_batch_size = suggestion_batch_size
    self.n_feature_dimensions_with_padding = n_feature_dimensions_with_padding
    self.n_feature_dimensions = n_feature_dimensions_with_padding
    self.dtype = types.ContinuousAndCategorical(jnp.float64, types.INT_DTYPE)

    self._categorical_logits = None
    if categorical_sizes:
      categorical_logits = np.zeros(
          [len(categorical_sizes), max(categorical_sizes)]
      )
      for i, s in enumerate(categorical_sizes):
        categorical_logits[i, s:] = -np.inf
      self._categorical_logits = categorical_logits

  def init_state(
      self,
      seed: jax.Array,
      n_parallel: int = 1,
      *,
      prior_features: Optional[vb.VectorizedOptimizerInput] = None,
      prior_rewards: Optional[types.Array] = None,
  ) -> None:
    del seed
    return

  def suggest(
      self,
      seed: jax.Array,
      state: None,
      n_parallel: int = 1,
  ) -> vb.VectorizedOptimizerInput:
    del state
    cont_seed, cat_seed = jax.random.split(seed)
    cont_data = jax.random.uniform(
        cont_seed,
        shape=(
            self._suggestion_batch_size,
            n_parallel,
            self.n_feature_dimensions_with_padding.continuous,
        ),
    )
    if self._categorical_logits is None:
      cat_data = jnp.zeros(
          [self._suggestion_batch_size, n_parallel, 0], dtype=jnp.int32
      )
    else:
      cat_data = tfd.Categorical(logits=self._categorical_logits).sample(
          (self._suggestion_batch_size, n_parallel), seed=cat_seed
      )
    return vb.VectorizedOptimizerInput(cont_data, cat_data)

  def suggestion_batch_size(self) -> int:
    return self._suggestion_batch_size

  def update(
      self,
      seed: jax.Array,
      state: None,
      batch_features: vb.VectorizedOptimizerInput,
      batch_rewards: types.Array,
  ) -> None:
    return


def random_strategy_factory(
    converter: converters.TrialToModelInputConverter,
    suggestion_batch_size: int,
) -> vb.VectorizedStrategy:
  """Creates a new vectorized strategy based on the Protocol."""
  return RandomVectorizedStrategy(
      converter=converter,
      suggestion_batch_size=suggestion_batch_size,
  )


def create_random_optimizer(
    converter: converters.TrialToModelInputConverter,
    max_evaluations: int,
    suggestion_batch_size: int,
) -> vb.VectorizedOptimizer:
  """Creates a random optimizer."""
  return vb.VectorizedOptimizerFactory(
      strategy_factory=random_strategy_factory,
      max_evaluations=max_evaluations,
      suggestion_batch_size=suggestion_batch_size,
  )(converter=converter)


def create_random_optimizer_factory(
    max_evaluations: int, suggestion_batch_size: int
) -> vb.VectorizedOptimizerFactory:
  """Creates a random optimizer factory."""
  return vb.VectorizedOptimizerFactory(
      strategy_factory=random_strategy_factory,
      max_evaluations=max_evaluations,
      suggestion_batch_size=suggestion_batch_size,
  )
